"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const chalk_1 = require("chalk");
const ora_1 = __importDefault(require("ora"));
const pluralize_1 = __importDefault(require("pluralize"));
const helpers_1 = require("yargs/helpers");
const yargs_1 = __importDefault(require("yargs/yargs"));
const config_1 = require("./config");
const events_1 = require("./events");
const smoker_1 = require("./smoker");
const util_1 = require("./util");
const error_1 = require("./error");
const BEHAVIOR_GROUP = 'Behavior:';
/**
 * Main entry point for the CLI script
 * @param args Raw command-line arguments array
 */
async function main(args) {
    const y = (0, yargs_1.default)(args);
    const result = await (0, util_1.readPackageJson)({ cwd: __dirname });
    if (!result) {
        throw new Error('midnight-smoker could not find its own package.json! bogus');
    }
    const { packageJson } = result;
    const { version, homepage } = packageJson;
    /**
     * These options can be specified more than once.
     *
     * Unfortunately, the typing gets hinky if we try to define the options object outside of the `options()` call, so we cannot programmatically gather these.
     */
    const arrayOptNames = new Set([
        'add',
        'workspace',
        'pm',
        'script',
        'scripts',
    ]);
    try {
        const config = await (0, config_1.readConfig)();
        if (config.script) {
            args.unshift(...config.script);
        }
        await y
            .scriptName('smoker')
            .command('* <script> [scripts..]', 'Run tests against a package as it would be published', (yargs) => {
            /**
             * Reusable config for array-type options
             */
            const arrayOptConfig = {
                requiresArg: true,
                array: true,
                type: 'string',
                coerce: util_1.normalizeStringArray,
            };
            return yargs
                .positional('script', {
                describe: 'Script in package.json to run',
                type: 'string',
                coerce: util_1.normalizeStringArray,
                default: config.script,
            })
                .positional('scripts', {
                describe: 'Additional script(s) to run',
                type: 'string',
                // no default here since if the config file is used, everything is
                // thrown into `script`.
                coerce: util_1.normalizeStringArray,
            })
                .options({
                add: {
                    describe: 'Additional dependency to provide to script(s)',
                    group: BEHAVIOR_GROUP,
                    default: config.add,
                    ...arrayOptConfig,
                },
                all: {
                    describe: 'Run script in all workspaces',
                    group: BEHAVIOR_GROUP,
                    default: config.all,
                    type: 'boolean',
                },
                bail: {
                    describe: 'When running scripts, halt on first error',
                    group: BEHAVIOR_GROUP,
                    default: config.bail,
                    type: 'boolean',
                },
                'include-root': {
                    describe: "Include the workspace root; must provide '--all'",
                    group: BEHAVIOR_GROUP,
                    default: config.includeRoot,
                    implies: 'all',
                    type: 'boolean',
                },
                json: {
                    describe: 'Output JSON only',
                    default: config.json,
                    group: BEHAVIOR_GROUP,
                    type: 'boolean',
                },
                linger: {
                    describe: 'Do not clean up temp dir(s) after completion',
                    group: BEHAVIOR_GROUP,
                    default: config.linger,
                    hidden: true,
                    type: 'boolean',
                },
                pm: {
                    describe: 'Run script(s) with a specific package manager; <npm|yarn|pnpm>[@version]',
                    group: BEHAVIOR_GROUP,
                    default: config.pm ?? 'npm@latest',
                    ...arrayOptConfig,
                },
                loose: {
                    describe: 'Ignore missing scripts (used with --all)',
                    type: 'boolean',
                    default: config.loose,
                    group: BEHAVIOR_GROUP,
                    implies: 'all',
                },
                workspace: {
                    describe: 'Run script in a specific workspace or workspaces',
                    group: BEHAVIOR_GROUP,
                    default: config.workspace,
                    ...arrayOptConfig,
                },
            })
                .check((argv) => {
                if (argv.pm?.some((pm) => pm.startsWith('pnpm'))) {
                    throw new error_1.SmokerError('pnpm is currently unsupported');
                }
                return true;
            });
        }, async (argv) => {
            const scripts = [...argv.script, ...argv.scripts];
            let smoker;
            try {
                smoker = await smoker_1.Smoker.init(scripts, argv);
            }
            catch (err) {
                console.error((0, chalk_1.red)(err.message));
                process.exitCode = 1;
                return;
            }
            if (argv.json) {
                let output;
                const setResult = (result) => {
                    smoker
                        .removeAllListeners(events_1.Events.RUN_SCRIPTS_OK)
                        .removeAllListeners(events_1.Events.RUN_SCRIPTS_FAILED);
                    output = result;
                };
                const cleanup = () => {
                    smoker
                        .removeAllListeners(events_1.Events.SMOKE_OK)
                        .removeAllListeners(events_1.Events.SMOKE_FAILED);
                };
                const writeOk = () => {
                    cleanup();
                    console.log(JSON.stringify(output, null, 2));
                };
                const writeFailed = () => {
                    cleanup();
                    console.log(JSON.stringify(output, null, 2));
                    process.exitCode = 1;
                };
                smoker
                    .once(events_1.Events.RUN_SCRIPTS_OK, setResult)
                    .once(events_1.Events.RUN_SCRIPTS_FAILED, setResult)
                    .once(events_1.Events.SMOKE_FAILED, writeFailed)
                    .once(events_1.Events.SMOKE_OK, writeOk);
                await smoker.smoke();
            }
            else {
                const spinner = (0, ora_1.default)();
                const scriptFailedEvts = [];
                smoker
                    .on(events_1.Events.SMOKE_BEGIN, () => {
                    console.error(`💨 ${(0, chalk_1.blue)('midnight-smoker')} ${(0, chalk_1.white)(`v${version}`)}`);
                })
                    .on(events_1.Events.PACK_BEGIN, () => {
                    let what;
                    if (argv.workspace?.length) {
                        what = (0, pluralize_1.default)('workspace', argv.workspace.length, true);
                    }
                    else if (argv.all) {
                        what = 'all workspaces';
                        if (argv.includeRoot) {
                            what += ' (and the workspace root)';
                        }
                    }
                    else {
                        what = 'current project';
                    }
                    spinner.start(`Packing ${what}...`);
                })
                    .on(events_1.Events.PACK_OK, ({ uniquePkgs, packageManagers }) => {
                    let msg = `Packed ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} using `;
                    if (packageManagers.length > 1) {
                        msg += `${(0, pluralize_1.default)('package manager', packageManagers.length, true)}`;
                    }
                    else {
                        msg += `${packageManagers[0]}`;
                    }
                    msg += '…';
                    spinner.succeed(msg);
                })
                    .on(events_1.Events.PACK_FAILED, (err) => {
                    spinner.fail(err.message);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.INSTALL_BEGIN, ({ uniquePkgs, packageManagers, additionalDeps }) => {
                    let msg = `Installing ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} from tarball`;
                    if (additionalDeps.length) {
                        msg += ` with ${(0, pluralize_1.default)('additional dependency', additionalDeps.length, true)}`;
                    }
                    if (packageManagers.length > 1) {
                        msg += ` using ${(0, pluralize_1.default)('package manager', packageManagers.length, true)}`;
                    }
                    else {
                        msg += ` using ${packageManagers[0]}`;
                    }
                    msg += '…';
                    spinner.start(msg);
                })
                    .on(events_1.Events.INSTALL_FAILED, (err) => {
                    spinner.fail(err.message);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.INSTALL_OK, ({ uniquePkgs }) => {
                    spinner.succeed(`Installed ${(0, pluralize_1.default)('unique package', uniquePkgs.length, true)} from tarball`);
                })
                    .on(events_1.Events.RUN_SCRIPTS_BEGIN, ({ total }) => {
                    spinner.start(`Running script 0/${total}...`);
                })
                    .on(events_1.Events.RUN_SCRIPT_BEGIN, ({ current, total }) => {
                    spinner.text = `Running script ${current}/${total}...`;
                })
                    .on(events_1.Events.RUN_SCRIPT_FAILED, (evt) => {
                    scriptFailedEvts.push(evt);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.RUN_SCRIPTS_OK, ({ total }) => {
                    spinner.succeed(`Successfully ran ${(0, pluralize_1.default)('script', total, true)}`);
                })
                    .on(events_1.Events.RUN_SCRIPTS_FAILED, ({ total, failures }) => {
                    spinner.fail(`${failures} of ${total} ${(0, pluralize_1.default)('script', total)} failed`);
                    for (const evt of scriptFailedEvts) {
                        console.error(`\n${(0, chalk_1.red)('Error details')} for failed package ${(0, chalk_1.yellow)(evt.pkgName)}:\n`);
                        console.error(evt.error.message);
                    }
                    process.exitCode = 1;
                })
                    .on(events_1.Events.SMOKE_FAILED, (err) => {
                    spinner.fail(err?.message ?? err);
                    process.exitCode = 1;
                })
                    .on(events_1.Events.SMOKE_OK, () => {
                    spinner.succeed('Lovey-dovey! 💖');
                })
                    .on(events_1.Events.LINGERED, (dirs) => {
                    console.error(`Lingering ${(0, pluralize_1.default)('temp directory', dirs.length)}:\n`);
                    for (const dir of dirs) {
                        console.error(`  ${(0, chalk_1.yellow)(dir)}`);
                    }
                });
                await smoker.smoke();
            }
        })
            .options({
            verbose: {
                describe: 'Verbose output',
                type: 'boolean',
                global: true,
            },
        })
            .epilog(`For more info, see ${homepage}\n`)
            .middleware(
        /**
         * If an array-type option is provided in both the config file and on the command-line,
         * we use this to merge the two arrays.
         *
         * If we had any object-type options (we don't) then we'd do that here as well.
         */
        (argv) => {
            for (const key of arrayOptNames) {
                const cfgKey = key;
                const arg = key;
                if (cfgKey in config && arg in argv) {
                    const cfgValue = (Array.isArray(config[cfgKey])
                        ? config[cfgKey]
                        : [config[cfgKey]]);
                    const argvValue = (Array.isArray(argv[arg]) ? argv[arg] : [argv[arg]]);
                    argv[arg] = [...new Set([...cfgValue, ...argvValue])];
                }
            }
            // squelch some output if `json` is true
            argv.verbose = argv.json ? false : argv.verbose;
        })
            .showHelpOnFail(false)
            .help()
            .strict()
            .parseAsync();
    }
    catch (err) {
        console.error((0, chalk_1.red)(err));
        process.exitCode = 1;
    }
}
main((0, helpers_1.hideBin)(process.argv));
//# sourceMappingURL=cli.js.map