/// <reference types="node" />
import { EventEmitter } from 'node:events';
import StrictEventEmitter from 'strict-event-emitter-types';
import { type SmokerEvents } from './events';
import { type PackageManager } from './pm';
import type { PkgInstallManifest, PkgRunManifest, RunScriptResult, SmokeOptions, SmokerOptions } from './types';
export declare const TMP_DIR_PREFIX = "midnight-smoker-";
type TSmokerEmitter = StrictEventEmitter<EventEmitter, SmokerEvents>;
declare const Smoker_base: new () => TSmokerEmitter;
export declare class Smoker extends Smoker_base {
    readonly pms: Map<string, PackageManager>;
    /**
     * List of extra dependencies to install
     */
    private readonly add;
    /**
     * Whether to run against all workspaces
     */
    private readonly allWorkspaces;
    /**
     * Whether to bail on the first script failure
     */
    private readonly bail;
    /**
     * Whether to include the workspace root
     */
    private readonly includeWorkspaceRoot;
    /**
     * Whether to keep temp dirs around (debugging purposes)
     */
    private readonly linger;
    /**
     * Mapping of {@linkcode PackageManager} instances to their identifiers of the form `<npm|yarn|pnpm>@<version>`
     */
    private readonly pmIds;
    /**
     * List of temp directories created
     */
    private readonly tempDirs;
    /**
     * List of specific workspaces to run against
     */
    private readonly workspaces;
    /**
     * List of scripts to run in each workspace
     */
    readonly scripts: string[];
    constructor(scripts: string | string[], pms: Map<string, PackageManager>, opts?: SmokerOptions);
    static init(scripts: string | string[], opts?: SmokeOptions): Promise<Smoker>;
    /**
     * Run the smoke test scripts!
     * @param scripts - One or more npm scripts to run
     * @param opts - Options
     */
    static smoke(scripts: string | string[], opts?: SmokeOptions): Promise<RunScriptResult[]>;
    /**
     * Cleans up any temp directories created by {@linkcode createTempDir}.
     *
     * If the {@linkcode SmokeOptions.linger} option is set to `true`, this method
     * will _not_ clean up the directories, but will instead emit a
     * {@linkcode SmokerEvents.Lingered|Lingered} event.
     */
    cleanup(): Promise<void>;
    /**
     * Creates a temp dir and adds it to the set of temp dirs to be cleaned up later
     * @returns New temp dir path
     */
    createTempDir(): Promise<string>;
    /**
     * Installs from tarball in a temp dir
     */
    install(manifests: PkgInstallManifest): Promise<PkgRunManifest>;
    /**
     * For each package manager, creates a tarball for one or more packages
     */
    pack(): Promise<PkgInstallManifest>;
    /**
     * Runs the script for each package in `packItems`
     */
    runScripts(pkgRunManifest: PkgRunManifest): Promise<RunScriptResult[]>;
    smoke(): Promise<RunScriptResult[]>;
    /**
     * This is only here because it's a fair amount of work to mash the data into a format more suitable for display.
     *
     * This is used by the events {@linkcode SmokerEvents.InstallBegin}, {@linkcode SmokerEvents.InstallOk}, and {@linkcode SmokerEvents.PackOk}.
     * @param pkgInstallManifest What to install and with what package manager
     * @returns Something to be emitted
     */
    private buildEventData;
}
export {};
//# sourceMappingURL=smoker.d.ts.map