"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Smoker = exports.TMP_DIR_PREFIX = void 0;
/* eslint-disable no-labels */
const node_events_1 = require("node:events");
const promises_1 = __importDefault(require("node:fs/promises"));
const node_os_1 = require("node:os");
const node_path_1 = __importDefault(require("node:path"));
const error_1 = require("./error");
const events_1 = require("./events");
const pm_1 = require("./pm");
const util_1 = require("./util");
exports.TMP_DIR_PREFIX = 'midnight-smoker-';
function createStrictEventEmitterClass() {
    const TypedEmitter = node_events_1.EventEmitter;
    return TypedEmitter;
}
const { SMOKE_BEGIN, SMOKE_OK, SMOKE_FAILED, PACK_BEGIN, PACK_FAILED, PACK_OK, INSTALL_BEGIN, INSTALL_FAILED, INSTALL_OK, RUN_SCRIPTS_BEGIN, RUN_SCRIPTS_FAILED, RUN_SCRIPTS_OK, RUN_SCRIPT_BEGIN, RUN_SCRIPT_FAILED, RUN_SCRIPT_OK, LINGERED, } = events_1.Events;
class Smoker extends createStrictEventEmitterClass() {
    constructor(scripts, pms, opts = {}) {
        super();
        this.pms = pms;
        this.scripts = (0, util_1.normalizeStringArray)(scripts);
        opts = { ...opts };
        this.linger = Boolean(opts.linger);
        this.includeWorkspaceRoot = Boolean(opts.includeRoot);
        if (this.includeWorkspaceRoot) {
            opts.all = true;
        }
        this.add = (0, util_1.normalizeStringArray)(opts.add);
        this.bail = Boolean(opts.bail);
        this.allWorkspaces = Boolean(opts.all);
        this.workspaces = (0, util_1.normalizeStringArray)(opts.workspace);
        if (this.allWorkspaces && this.workspaces.length) {
            throw new error_1.SmokerError('Option "workspace" is mutually exclusive with "all" and/or "includeRoot"');
        }
        this.pmIds = new WeakMap();
        for (const [pmId, pm] of pms) {
            this.pmIds.set(pm, pmId);
        }
        this.tempDirs = new Set();
    }
    static async init(scripts, opts = {}) {
        const { pm, verbose, loose, all } = opts;
        const pms = await (0, pm_1.loadPackageManagers)(pm, {
            verbose,
            loose: all && loose,
        });
        return new Smoker(scripts, pms, opts);
    }
    /**
     * Run the smoke test scripts!
     * @param scripts - One or more npm scripts to run
     * @param opts - Options
     */
    static async smoke(scripts, opts = {}) {
        const smoker = await Smoker.init(scripts, opts);
        return smoker.smoke();
    }
    /**
     * Cleans up any temp directories created by {@linkcode createTempDir}.
     *
     * If the {@linkcode SmokeOptions.linger} option is set to `true`, this method
     * will _not_ clean up the directories, but will instead emit a
     * {@linkcode SmokerEvents.Lingered|Lingered} event.
     */
    async cleanup() {
        if (!this.linger) {
            await Promise.all([...this.tempDirs].map(async (tempdir) => {
                try {
                    await promises_1.default.rm(tempdir, { recursive: true, force: true });
                }
                catch (e) {
                    const err = e;
                    if (err.code !== 'ENOENT') {
                        throw new error_1.SmokerError(`Failed to clean temp directory ${tempdir}: ${e}`);
                    }
                }
                finally {
                    this.tempDirs.delete(tempdir);
                }
            }));
        }
        else if (this.tempDirs.size) {
            await Promise.resolve();
            this.emit(LINGERED, [...this.tempDirs]);
        }
    }
    /**
     * Creates a temp dir and adds it to the set of temp dirs to be cleaned up later
     * @returns New temp dir path
     */
    async createTempDir() {
        // TODO EMIT
        try {
            const prefix = node_path_1.default.join((0, node_os_1.tmpdir)(), exports.TMP_DIR_PREFIX);
            const tempdir = await promises_1.default.mkdtemp(prefix);
            this.tempDirs.add(tempdir);
            return tempdir;
        }
        catch (err) {
            throw new error_1.SmokerError(`Failed to create temporary directory: ${err}`);
        }
    }
    /**
     * Installs from tarball in a temp dir
     */
    async install(manifests) {
        if (!manifests?.size) {
            throw new TypeError('(install) Non-empty "pkgInstallManifest" arg is required');
        }
        // ensure we emit asynchronously
        await Promise.resolve();
        const installData = this.buildEventData(manifests);
        this.emit(INSTALL_BEGIN, installData);
        const pkgRunManifest = new Map();
        for (const [pm, manifest] of manifests) {
            try {
                // TODO check for errors?
                await pm.install({ ...manifest, additionalDeps: this.add });
                const runManifests = new Set();
                for (const packedPkg of manifest.packedPkgs) {
                    for (const script of this.scripts) {
                        runManifests.add({ packedPkg, script });
                    }
                }
                pkgRunManifest.set(pm, runManifests);
            }
            catch (err) {
                const error = err;
                this.emit(INSTALL_FAILED, error);
                throw error;
            }
        }
        this.emit(INSTALL_OK, installData);
        return pkgRunManifest;
    }
    /**
     * For each package manager, creates a tarball for one or more packages
     */
    async pack() {
        await Promise.resolve();
        this.emit(PACK_BEGIN);
        const manifestMap = new Map();
        for await (const pm of this.pms.values()) {
            const dest = await this.createTempDir();
            let manifest;
            try {
                manifest = await pm.pack(dest, {
                    allWorkspaces: this.allWorkspaces,
                    includeWorkspaceRoot: this.includeWorkspaceRoot,
                    workspaces: this.workspaces,
                });
                manifestMap.set(pm, manifest);
            }
            catch (err) {
                this.emit(PACK_FAILED, err);
                throw err;
            }
        }
        this.emit(PACK_OK, this.buildEventData(manifestMap));
        return manifestMap;
    }
    /**
     * Runs the script for each package in `packItems`
     */
    async runScripts(pkgRunManifest) {
        if (!pkgRunManifest) {
            throw new TypeError('(runScripts) "pkgRunManifest" arg is required');
        }
        const pkgRunManifestForEmit = {};
        for (const [pm, runManifests] of pkgRunManifest) {
            const pmId = this.pmIds.get(pm);
            if (!pmId) {
                /* istanbul ignore next */
                throw new error_1.SmokerError('Could not find package manager ID; please report this bug');
            }
            pkgRunManifestForEmit[pmId] = [...runManifests];
        }
        const totalScripts = [...pkgRunManifest].reduce((count, [, runManifests]) => count + runManifests.size, 0);
        await Promise.resolve();
        this.emit(RUN_SCRIPTS_BEGIN, {
            manifest: pkgRunManifestForEmit,
            total: totalScripts,
        });
        const { bail } = this;
        let current = 0;
        const scripts = [];
        const results = [];
        BAIL: for await (const [pm, runManifests] of pkgRunManifest) {
            const pmId = this.pmIds.get(pm);
            if (!pmId) {
                /* istanbul ignore next */
                throw new error_1.SmokerError('Could not find package manager ID; please report this bug');
            }
            for await (const runManifest of runManifests) {
                let result;
                const { script } = runManifest;
                scripts.push(script);
                this.emit(RUN_SCRIPT_BEGIN, {
                    script,
                    pkgName: runManifest.packedPkg.pkgName,
                    total: totalScripts,
                    current,
                });
                try {
                    result = await pm.runScript(runManifest);
                    results.push(result);
                    if (result.error) {
                        this.emit(RUN_SCRIPT_FAILED, {
                            ...result,
                            script,
                            error: result.error,
                            current: current++,
                            total: totalScripts,
                        });
                        if (bail) {
                            break BAIL;
                        }
                    }
                    else {
                        this.emit(RUN_SCRIPT_OK, {
                            ...result,
                            script,
                            current,
                            total: totalScripts,
                        });
                    }
                }
                catch (err) {
                    throw new error_1.SmokerError(`(runScripts): Unknown failure from "${pmId}" plugin: ${err}`);
                }
            }
        }
        const failureCount = results.filter((result) => result.error).length;
        if (failureCount) {
            this.emit(RUN_SCRIPTS_FAILED, {
                results,
                manifest: pkgRunManifestForEmit,
                total: totalScripts,
                failures: failureCount,
                executed: results.length,
            });
        }
        else {
            this.emit(RUN_SCRIPTS_OK, {
                failures: 0,
                results,
                manifest: pkgRunManifestForEmit,
                total: totalScripts,
                executed: results.length,
            });
        }
        return results;
    }
    async smoke() {
        // do not emit synchronously
        await Promise.resolve();
        this.emit(SMOKE_BEGIN);
        try {
            const pkgInstallManifest = await this.pack();
            const pkgRunManifest = await this.install(pkgInstallManifest);
            const results = await this.runScripts(pkgRunManifest);
            if (!results.some((result) => result.error)) {
                this.emit(SMOKE_OK);
            }
            else {
                this.emit(SMOKE_FAILED, new error_1.SmokerError('🤮 Maurice!'));
            }
            return results;
        }
        catch (err) {
            this.emit(SMOKE_FAILED, err);
            throw err;
        }
        finally {
            await this.cleanup();
        }
    }
    /**
     * This is only here because it's a fair amount of work to mash the data into a format more suitable for display.
     *
     * This is used by the events {@linkcode SmokerEvents.InstallBegin}, {@linkcode SmokerEvents.InstallOk}, and {@linkcode SmokerEvents.PackOk}.
     * @param pkgInstallManifest What to install and with what package manager
     * @returns Something to be emitted
     */
    buildEventData(pkgInstallManifest) {
        const uniquePkgs = new Set();
        const pmIds = new Set();
        const manifests = [];
        const additionalDeps = new Set();
        for (const [pm, manifest] of pkgInstallManifest) {
            const id = this.pmIds.get(pm);
            if (!id) {
                /* istanbul ignore next */
                throw new error_1.SmokerError('Could not find package manager ID; please report this bug');
            }
            pmIds.add(id);
            manifests.push(manifest);
            for (const { pkgName } of manifest.packedPkgs) {
                uniquePkgs.add(pkgName);
            }
            if (manifest.additionalDeps) {
                for (const dep of manifest.additionalDeps) {
                    additionalDeps.add(dep);
                }
            }
        }
        return {
            uniquePkgs: [...uniquePkgs],
            packageManagers: [...pmIds],
            additionalDeps: [...additionalDeps],
            manifests,
        };
    }
}
exports.Smoker = Smoker;
//# sourceMappingURL=smoker.js.map